package config

import (
	"bufio"
	"fmt"
	"io/fs"
	"io/ioutil"
	"os"
	"path"
	"path/filepath"
	"strings"

	"github.com/Autonomic-Cooperative/godotenv"
	"github.com/sirupsen/logrus"
)

var ABRA_DIR = os.ExpandEnv("$HOME/.abra")
var SERVERS_DIR = path.Join(ABRA_DIR, "servers")
var RECIPES_DIR = path.Join(ABRA_DIR, "recipes")
var VENDOR_DIR = path.Join(ABRA_DIR, "vendor")
var BACKUP_DIR = path.Join(ABRA_DIR, "backups")
var RECIPES_JSON = path.Join(ABRA_DIR, "catalogue", "recipes.json")
var REPOS_BASE_URL = "https://git.coopcloud.tech/coop-cloud"
var CATALOGUE_JSON_REPO_NAME = "recipes-catalogue-json"
var SSH_URL_TEMPLATE = "ssh://git@git.coopcloud.tech:2222/coop-cloud/%s.git"

// GetServers retrieves all servers.
func GetServers() ([]string, error) {
	var servers []string

	servers, err := GetAllFoldersInDirectory(SERVERS_DIR)
	if err != nil {
		return servers, err
	}

	logrus.Debugf("retrieved %v servers: %s", len(servers), servers)

	return servers, nil
}

// ReadEnv loads an app envivornment into a map.
func ReadEnv(filePath string) (AppEnv, error) {
	var envFile AppEnv

	envFile, err := godotenv.Read(filePath)
	if err != nil {
		return nil, err
	}

	logrus.Debugf("read %s from %s", envFile, filePath)

	return envFile, nil
}

// ReadServerNames retrieves all server names.
func ReadServerNames() ([]string, error) {
	serverNames, err := GetAllFoldersInDirectory(SERVERS_DIR)

	if err != nil {
		return nil, err
	}

	logrus.Debugf("read %s from %s", strings.Join(serverNames, ","), SERVERS_DIR)

	return serverNames, nil
}

// getAllFilesInDirectory returns filenames of all files in directory
func getAllFilesInDirectory(directory string) ([]fs.FileInfo, error) {
	var realFiles []fs.FileInfo

	files, err := ioutil.ReadDir(directory)
	if err != nil {
		return nil, err
	}

	for _, file := range files {
		// Follow any symlinks
		filePath := path.Join(directory, file.Name())

		if filepath.Ext(strings.TrimSpace(filePath)) != ".env" {
			continue
		}

		realPath, err := filepath.EvalSymlinks(filePath)
		if err != nil {
			logrus.Warningf("broken symlink in your abra config folders: %s", filePath)
		} else {
			realFile, err := os.Stat(realPath)
			if err != nil {
				return nil, err
			}
			if !realFile.IsDir() {
				realFiles = append(realFiles, file)
			}
		}
	}

	return realFiles, nil
}

// GetAllFoldersInDirectory returns both folder and symlink paths
func GetAllFoldersInDirectory(directory string) ([]string, error) {
	var folders []string

	files, err := ioutil.ReadDir(directory)
	if err != nil {
		return nil, err
	}
	if len(files) == 0 {
		return nil, fmt.Errorf("directory is empty: %s", directory)
	}

	for _, file := range files {
		// Check if file is directory or symlink
		if file.IsDir() || file.Mode()&fs.ModeSymlink != 0 {
			filePath := path.Join(directory, file.Name())
			realDir, err := filepath.EvalSymlinks(filePath)
			if err != nil {
				logrus.Warningf("broken symlink in your abra config folders: %s", filePath)
			} else if stat, err := os.Stat(realDir); err == nil && stat.IsDir() {
				// path is a directory
				folders = append(folders, file.Name())
			}
		}
	}

	return folders, nil
}

// ReadAbraShEnvVars reads env vars from an abra.sh recipe file.
func ReadAbraShEnvVars(abraSh string) (map[string]string, error) {
	envVars := make(map[string]string)

	file, err := os.Open(abraSh)
	if err != nil {
		if os.IsNotExist(err) {
			return envVars, nil
		}
		return envVars, err
	}

	scanner := bufio.NewScanner(file)
	for scanner.Scan() {
		line := scanner.Text()
		if strings.Contains(line, "export") {
			splitVals := strings.Split(line, "export ")
			envVarDef := splitVals[len(splitVals)-1]
			keyVal := strings.Split(envVarDef, "=")
			if len(keyVal) != 2 {
				return envVars, fmt.Errorf("couldn't parse %s", line)
			}
			envVars[keyVal[0]] = keyVal[1]
		}
	}

	logrus.Debugf("read %s from %s", envVars, abraSh)

	return envVars, nil
}
